<?php

/**
 * @file
 * Contains Drupal\maestro\Form\MaestroTemplateFormBase.
 */

namespace Drupal\maestro\Form;

use Drupal\Core\Entity\EntityFormController;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\maestro\Engine\MaestroEngine;

/**
 * Class MaestroTemplateFormBase.
 *
 *
 * @package Drupal\maestro\Form
 *
 * @ingroup maestro
 */
class MaestroTemplateFormBase extends EntityForm {

  /**
   * Overrides Drupal\Core\Entity\EntityFormController::form().
   *
   * Builds the entity add/edit form.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param array $form_state
   *
   * @return array
   *   An associative array containing the Template add/edit form.
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Get anything we need form the base class.
    $form = parent::buildForm($form, $form_state);

    $Template = $this->entity;

    // Build the form.
    $form['label'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#maxlength' => 255,
      '#default_value' => $Template->label(),
      '#required' => TRUE,
    );
    
    $form['id'] = array(
      '#type' => 'machine_name',
      '#title' => $this->t('Machine name'),
      '#default_value' => $Template->id(),
      '#machine_name' => array(
        'exists' =>  [get_class($this), 'exists'],
      ),
      //'#disabled' => !$Template->isNew(),  //I believe we should keep the machine name editable
    );

    //The notion of App Groups will be carried across to D8 Maestro, however we
    //will simply hide for the time being as they were used sparingly in the Drupal 7 version
    $form['app_group'] = array(
      '#type' => 'hidden',
      '#title' => $this->t('App Group'),
      '#maxlength' => 255,
      '#default_value' => isset($Template->app_group) ? $Template->app_group : 0,
      '#required' => TRUE,
    );

    //Canvas height is a default value of 900 pixels.
    $form['canvas_height'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Canvas Height in pixels'),
      '#maxlength' => 255,
      '#default_value' => isset($Template->canvas_height) ? $Template->canvas_height : 900 ,
      '#required' => TRUE,
    );
    
    //default is 800 px wide
    $form['canvas_width'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Canvas Width in pixels'),
      '#maxlength' => 255,
      '#default_value' => isset($Template->canvas_width) ? $Template->canvas_width : 800 ,
      '#required' => TRUE,
    );

   
    //variable definitions:
    //Complex type of machine name and variable value
    $form['variables'] = array(
      '#type' => 'details',
      '#title' => 'Variables',
      '#group' => 'templatevariables',
      '#tree' => TRUE,
      '#prefix' => '<div id="multi-replace">',
      '#suffix' => '</div>'
    );

    $key = 0;
    if(isset($this->entity->variables)) {
      foreach($this->entity->variables as $key => $variable) {
        $disabled = FALSE;
        $attributes = NULL;
        if($key == 'initiator' || $key == 'entity_identifiers') {
          //do not allow the initiator or entity_identifiers to be deleted
          $disabled = TRUE;
          $attributes = array(
            'title' => t("You cannot delete this variable.  This variable is required by Maestro."),
          );
        }
        
        $form['variables'][$key]['delete'] = array(
          '#type' => 'checkbox',
          '#title' => $this->t('Delete'),
          '#title_display' => 'before',
          '#required' => FALSE,
          '#prefix' => '<div class="clearfix maestro-existing-variable">',
          '#disabled' => $disabled,
          '#attributes' => $attributes,
        );
        
        $form['variables'][$key]['variable_id'] = array(
          '#type' => 'machine_name',
          '#title' => $this->t('Variable name'),
          '#default_value' => $this->entity->variables[$key]['variable_id'],
          '#machine_name' => array(
              'exists' =>  [get_class($this), 'exists'],
          ),
          '#required' => FALSE,
        );
        
        $form['variables'][$key]['variable_value'] = array(
          '#type' => 'textfield',
          '#title' => $this->t('Value'),
          '#maxlength' => 255,
          '#default_value' => $this->entity->variables[$key]['variable_value'],
          '#required' => FALSE,
          '#suffix' => '</div>',
        );
      }
    }
    
    
    //form elements for a new variable
    $form['variables']['new_fieldset'] = array(
      '#type' => 'fieldset',
      '#title' => 'Create New Variable'
    );
    
    $form['variables']['new_fieldset']['new']['variable_id'] = array(
      '#type' => 'machine_name',
      '#title' => $this->t('Variable name'),
      '#default_value' => '',
      '#machine_name' => array(
          'exists' =>  [get_class($this), 'exists'],
      ),
      '#required' => FALSE,
    );
    
    $form['variables']['new_fieldset']['new']['variable_value'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Value'),
      '#maxlength' => 255,
      '#default_value' => '',
      '#required' => FALSE,
    );
    
    $form['#attached']['library'][] = 'maestro/maestro-engine-css';
    return $form;
  }

  
  static public function exists($submitted_value, array $element, FormStateInterface $form_state) {
    $templates = MaestroEngine::getTemplates();
    if(array_key_exists($submitted_value, $templates)) {
      return TRUE;
    }
    return FALSE;
  }
  
  /**
   * Overrides Drupal\Core\Entity\EntityFormController::actions().
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param array $form_state
   *
   * @return array
   *   An array of supported actions for the current entity form.
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    $actions = parent::actions($form, $form_state);
    $actions['submit']['#value'] = $this->t('Save');
    return $actions;
  }

  /**
   * Overrides Drupal\Core\Entity\EntityFormController::validate().
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param array $form_state
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    parent::validateForm($form, $form_state);
  }

  /**
   * Overrides Drupal\Core\Entity\EntityFormController::save().
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param array $form_state
   */
  public function save(array $form, FormStateInterface $form_state) {
    //remove the 'new' variable values here so that the save, well, saves them.
    //form_state->values['variables'] holds the saved variables
    //we need to ensure that the variables at this stage, since the parent validation has happened
    //is properly formatted to remove the 'new' key and add that to the numerical keys
    
    $variables = $form_state->getValue('variables');
    $new = $variables['new_fieldset']['new'];
    unset($variables['new_fieldset']);
    if(isset($new['variable_id']) && !empty($new['variable_id'])) {
      //we have a new variable here.
      $variables[$new['variable_id']] = $new;
    }
    
    //now to handle deletion
    foreach($variables as $key => $variable) {
      if(array_key_exists('delete', $variable) && isset($variable['delete'])) {
        if($variable['delete'] == 1) unset($variables[$key]);
        else unset($variables[$key]['delete']);
      }
    }
    
    $this->entity->variables = $variables;
    $status = $this->entity->save();

    if ($status == SAVED_UPDATED) {
      // If we edited an existing entity...
      drupal_set_message($this->t('Template %label has been updated.', array('%label' => $this->entity->label())));
      \Drupal::logger('contact')->notice('Template %label has been updated.', array('%label' => $this->entity->label()));
    }
    else {
      // If we created a new entity...
      drupal_set_message($this->t('Template %label has been added.', array('%label' => $this->entity->label())));
      \Drupal::logger('contact')->notice('Template %label has been added.', array('%label' => $this->entity->label()));
      $form_state->setRedirect('entity.maestro_template.list');
    }

    
  }

}
