<?php

namespace Drupal\maestro\Plugin\EngineTasks;

use Drupal\Core\Plugin\PluginBase;
use Drupal\maestro\MaestroEngineTaskInterface;
use Drupal\maestro\MaestroTaskTrait;
use Drupal\maestro\Engine\MaestroEngine;
use Drupal\Core\Form\FormStateInterface;
use Drupal\maestro\Form\MaestroExecuteInteractive;

/**
 * Maestro Interactive Task Plugin.
 *
 * The plugin annotations below should include:
 * id: The task type ID for this task.  For Maestro tasks, this is Maestro[TaskType].
 *     So for example, the start task shipped by Maestro is MaestroStart.
 *     The Maestro End task has an id of MaestroEnd
 *     Those task IDs are what's used in the engine when a task is injected into the queue
 *
 * @Plugin(
 *   id = "MaestroInteractive",
 *   task_description = @Translation("The Maestro Engine's interactive task."),
 * )
 */
class MaestroInteractiveTask extends PluginBase implements MaestroEngineTaskInterface {

  use MaestroTaskTrait;
   
  function __construct($configuration = NULL) {
    if(is_array($configuration)) {
      $this->processID = $configuration[0];
      $this->queueID = $configuration[1];
    }
  }
  
  /**
   * {@inheritDoc}
   */
  public function isInteractive() {
    return TRUE;
  }
  
  /**
   * {@inheritDoc}
   */
  public function shortDescription() {
    return t('Interactive Task');
  }
  
  /**
   * {@inheritDoc}
   */
  public function description() {
    return $this->t('Interactive Task.');
  }

  /**
   * {@inheritDoc}
   * @see \Drupal\Component\Plugin\PluginBase::getPluginId()
   */
  public function getPluginId() {
    return 'MaestroInteractive';
  }
  
  /**
   * {@inheritDoc}
   */
  public function getTaskColours() {
    return '#0000ff';
  }
  
  /*
   * Part of the ExecutableInterface
   * Execution of the interactive task does nothing except for setting the run_once flag
   * {@inheritdoc}
   */
  public function execute() {
    //need to set the run_once flag here
    //as interactive tasks are executed and completed by the user using the Maestro API
    $queueRecord = \Drupal::entityTypeManager()->getStorage('maestro_queue')->load($this->queueID);
    $queueRecord->set('run_once', 1);
    $queueRecord->save();
    
    return FALSE;
  }
  
  /**
   * {@inheritDoc}
   */
  public function getExecutableForm($modal, MaestroExecuteInteractive $parent) {
    //By default, we will provide a form that has a base queueID field and a submit/reject button set.
    //you can override this with ease in your own handler.
    //the task console should be looking at whether this requires a handler form or not.
    //our implementation forces the handler to be a mandatory field, thus causing functions like: form maestro_accept_only_form in the maestro.module
    //file to fire.
    
    $form['queueID'] = array(   //we are hiding this for now, as you can override all of this with your own handler.  T
      '#type' => 'hidden',      //this is just a placeholder form to get you under way.
      '#title' => $this->t('The queue ID of this task'),
      '#default_value' => $this->queueID,
      '#description' => $this->t('queueID'),
    );
  
    $form['actions']['submit'] = array(
      '#type' => 'submit',
      '#value' => $this->t('Complete'),
    );
    
    $form['actions']['reject'] = array(
      '#type' => 'submit',
      '#value' => $this->t('Reject'),
    );
  
    if($modal == 'modal') {
      $form['actions']['submit']['#ajax'] = array(
        'callback' => [$parent, 'completeForm'],
        'wrapper' => '',
      );
      
      $form['actions']['reject']['#ajax'] = array(
        'callback' => [$parent, 'completeForm'],
        'wrapper' => '',
      );
    }
    return $form;
  }
  
  /**
   * {@inheritDoc}
   */
  public function handleExecuteSubmit(array &$form, FormStateInterface $form_state) {
  

    $queueID = intval($form_state->getValue('maestro_queue_id'));
    $triggeringElement = $form_state->getTriggeringElement();
    if(strstr($triggeringElement['#id'], 'edit-submit') !== FALSE && $queueID > 0) {
      MaestroEngine::completeTask($queueID, \Drupal::currentUser()->id());
    }
    else {
      //we'll complete the task, but we'll also flag it as TASK_STATUS_CANCEL
      MaestroEngine::completeTask($queueID, \Drupal::currentUser()->id());
      MaestroEngine::setTaskStatus($queueID, TASK_STATUS_CANCEL);
    }
    
   
  }
  
  /**
   * {@inheritDoc}
   */
  public function getTaskEditForm(array $task, $templateMachineName) {
    $form = array(
      '#markup' => t('Interactive Task Edit'),
    );
    
    $form['handler'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Handler'),
      '#description' => $this->t('The function that contains the form definition for this instance of the interactive task.'),
      '#default_value' => $task['handler'],
      '#required' => FALSE,
    );
    
    $form['reditect_to'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Return Path'),
      '#description' => $this->t('You can specify where your return path should go upon task completion.'),
      '#default_value' => isset($task['data']['reditect_to']) ? $task['data']['reditect_to'] : 'taskconsole',
      '#required' => TRUE,
    );
    
    $form['modal'] = array(
        '#type' => 'select',
        '#title' => $this->t('Task presentation'),
        '#description' => $this->t('Should this task be shown as a modal or full screen task.'),
        '#default_value' => $task['data']['modal'],
        '#options' => array(
          'modal' => $this->t('Modal'),
          'notmodal' => $this->t('Full Page'),
        ),
        '#required' => TRUE,
    );
    return $form;
  }
  
  /**
   * {@inheritDoc}
   */
  public function validateTaskEditForm(array &$form, FormStateInterface $form_state) {
    $handler = $form_state->getValue('handler');
    //Let's validate the handler here to ensure that it actually exists.
    if($handler != '' && !function_exists($handler)) {
      $form_state->setErrorByName('handler', $this->t('This handler form function does not exist.'));
    }
  }
  
  /**
   * {@inheritDoc}
   */
  public function prepareTaskForSave(array &$form, FormStateInterface $form_state, array &$task) {
    $task['handler'] = $form_state->getValue('handler');
    $task['data']['modal'] = $form_state->getValue('modal');
    $redirect = $form_state->getValue('reditect_to');
    if(isset($redirect)) {
      $task['data']['reditect_to'] = $redirect;
    }
    else {
      $task['data']['reditect_to'] = '';
    }
    
  }
  
  /**
   * {@inheritDoc}
   */
  public function performValidityCheck(array &$validation_failure_tasks, array &$validation_information_tasks, array $task) {
    //pretty simple -- just ensure that there's a handler and modal set
    if( (array_key_exists('handler', $task) && $task['handler'] == '')  || !array_key_exists('handler', $task)) {
      $validation_information_tasks[] = array(
        'taskID' => $task['id'],
        'taskLabel' => $task['label'],
        'reason' => t('The Interactive Task handler is missing and thus the engine will assign the default handler to this task.'),
      );
    }
    
    if( (array_key_exists('modal', $task['data']) && $task['data']['modal'] == '')  || !array_key_exists('modal', $task['data'])) {
      $validation_failure_tasks[] = array(
        'taskID' => $task['id'],
        'taskLabel' => $task['label'],
        'reason' => t('The Interactive Task has not been set up properly.  The "Task Presentation" option is missing and thus the engine will be unable to execute this task.'),
      );
    }
    
    //This task should have assigned users
    //$task['assigned'] should have data
    if( (array_key_exists('assigned', $task) && $task['assigned'] == '')  || !array_key_exists('assigned', $task)) {
      $validation_failure_tasks[] = array(
        'taskID' => $task['id'],
        'taskLabel' => $task['label'],
        'reason' => t('The Interactive Task has not been set up properly.  The Interactive Task requires assignments to actors, roles or other assignment options.'),
      );
    }
  }
  
  /**
   * {@inheritDoc}
   */
  public function getTemplateBuilderCapabilities() {
    return array('edit', 'drawlineto', 'removelines', 'remove');
  }
}
